package io.tswf.transmutations.staticview.generation

import groovy.transform.CompileStatic
import io.tswf.transmutations.staticview.exception.StaticViewInjectionException
import io.tswf.transmutations.staticview.exception.StaticViewMethodInjectionException
import io.tswf.transmutations.staticview.meta.StaticViewMetadata
import io.tswf.transmutations.common.conflict.ConflictResolver
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.VarInsnNode

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class StaticViewInjector {

    static void injectStaticView(StaticViewMetadata viewMeta, ClassNode targetClass) {
        if (targetClass.name != viewMeta.targetType.internalName) {
            throw new IllegalArgumentException("Static view type mismatch. Expected ${viewMeta.targetType.className}, but was ${targetClass.javaName}!")
        }

        try {
            for (def viewMethodMeta : viewMeta.viewMethodsMetadata) {
                try {
                    def accessorMethod = StaticAccessMethodGenerator.generateStaticAccessor(viewMethodMeta)
                    def viewMethod = viewMethodMeta.viewMethod

                    def conflictingMethodInTarget = targetClass.findSameMethod(accessorMethod)

                    if (conflictingMethodInTarget != null) {
                        def overwriteMethod = ConflictResolver.resolveConflict(
                                targetClass: targetClass,
                                conflictingMethod: conflictingMethodInTarget,
                                anotherMethodOwner: viewMeta.viewClass,
                                anotherMethod: accessorMethod,
                                strategyProvider: viewMethodMeta
                        )

                        if (!overwriteMethod) {
                            continue
                        }

                        targetClass.methods.remove(conflictingMethodInTarget)
                    }

                    targetClass.methods << accessorMethod

                    def viewMethodIsStatic = Modifier.isStatic(viewMethod.access)

                    viewMethod.instructions.clear()

                    if (!viewMethodIsStatic) {
                        viewMethod.instructions << new VarInsnNode(0, ALOAD)
                    }

                    viewMethod.parameterTypes.eachWithIndex{ parameterType, parameterIndex ->
                        def localVarIndex = parameterIndex
                        if (!viewMethodIsStatic) {
                            localVarIndex ++
                        }

                        viewMethod.instructions << new VarInsnNode(localVarIndex, parameterType.getOpcode(ILOAD))
                    }

                    viewMethod.instructions << new MethodInsnNode(INVOKESTATIC, targetClass.name, accessorMethod.name, accessorMethod.desc, false)
                    viewMethod.instructions << new InsnNode(viewMethod.returnType.getOpcode(IRETURN))

                    def viewMethodMappingContext = new StaticViewMethodMappingContext(
                            viewMethodMeta: viewMethodMeta,
                            targetClass: targetClass,
                            generatedAccessMethodInTarget: accessorMethod
                    )

                    viewMethodMeta.mapper.handleStaticViewMethodMapping(viewMethodMappingContext)
                } catch (Throwable e) {
                    throw new StaticViewMethodInjectionException(e, viewMethodMeta, targetClass)
                }
            }
        } catch (Throwable e) {
            throw new StaticViewInjectionException(e, viewMeta, targetClass)
        }
    }

}
